from .command import command
from .differ import DifferChanges, DifferCoverage, DifferDiscrepancy
from .generator import Generator
from .models import APIMethod, APIMethods, BodySchema, RawSpecMethod, ResponseBodySchema
from .parser import Parser

__all__ = (
    "command",
    "DifferChanges",
    "DifferCoverage",
    "DifferDiscrepancy",
    "Generator",
    "APIMethod",
    "APIMethods",
    "APIMethods",
    "BodySchema",
    "RawSpecMethod",
    "ResponseBodySchema",
    "Parser",
)



from dataclasses import dataclass
from typing import Any, Dict, Set

from schemax import SchemaData


@dataclass
class RawSpecMethod:
    method_spec_id: str
    method: str
    route: str
    query_params: Set[str]
    request_body_schema: Dict[str, Any]
    response_code: str
    response_body_schema: Dict[str, Any]

    @staticmethod
    def create(data: Any) -> "RawSpecMethod":
        if isinstance(data, SchemaData):
            return RawSpecMethod(
                method_spec_id=data.http_method + data.path,
                method=data.http_method,
                route=data.path,
                query_params=set(data.queries),
                request_body_schema=data.request_schema,
                response_code=str(data.status),
                response_body_schema=data.response_schema
            )
        else:
            raise ValueError("Unsupported data format")


@dataclass
class BodySchema:
    schema: Dict[str, Dict[str, Any]]

    @classmethod
    def create(cls, content_type: str, schema: Dict[str, Any]) -> "BodySchema":
        return BodySchema({
            content_type: schema
        })

    def get_content_types(self) -> Set[str]:
        return set(self.schema.keys())

    def __getitem__(self, key: str) -> Dict[str, Any]:
        return self.schema[key]


@dataclass
class ResponseBodySchema:
    schema: Dict[str, BodySchema]

    @classmethod
    def create(cls, status: str, content: BodySchema) -> "ResponseBodySchema":
        return ResponseBodySchema({
            status: content
        })

    def add_info(self, code: str, content_type: str, schema: Dict[str, Any]) -> None:
        if code not in schema:
            self.schema[code] = BodySchema.create(content_type, schema)

    def get_codes(self) -> Set[str]:
        return set(self.schema.keys())

    def __getitem__(self, key: str) -> BodySchema:
        return self.schema[key]


@dataclass
class APIMethod:
    method: str
    route: str
    query_params: Set[str]

    request_body_schema: BodySchema
    response_body_schema: ResponseBodySchema

    @classmethod
    def create(cls, rsm: RawSpecMethod) -> "APIMethod":
        return APIMethod(
            rsm.method,
            rsm.route,
            rsm.query_params,
            BodySchema.create("application/json", rsm.request_body_schema),
            ResponseBodySchema.create(
                rsm.response_code,
                BodySchema.create("application/json", rsm.response_body_schema)
            )
        )

    def add_info(self, spec_method: RawSpecMethod) -> None:
        self.response_body_schema.add_info(
            spec_method.response_code, "application/json", spec_method.response_body_schema
        )

    def get_codes(self) -> Set[str]:
        return self.response_body_schema.get_codes()


@dataclass
class APIMethods:
    methods: Dict[str, APIMethod]

    def add_method(self, rsm: RawSpecMethod) -> None:
        if rsm.method_spec_id not in self.methods:
            self.methods[rsm.method_spec_id] = APIMethod.create(rsm)
        else:
            self.methods[rsm.method_spec_id].add_info(rsm)

    def get_ids(self) -> Set[str]:
        return set(self.methods.keys())

    def __getitem__(self, key: str) -> APIMethod:
        return self.methods[key]



import requests
import yaml
from schemax import collect_schema_data

from .models import APIMethods, RawSpecMethod


class Parser:
    @classmethod
    def parse(cls, spec_path: str) -> APIMethods:
        if spec_path.startswith("http://") or spec_path.startswith("https://"):
            return cls.parse_from_url(spec_path)
        else:
            return cls.parse_from_file(spec_path)

    @staticmethod
    def parse_from_raw(content: str) -> APIMethods:
        data = yaml.load(content, Loader=yaml.CLoader)
        api_methods = APIMethods({})
        for item in collect_schema_data(data):
            api_methods.add_method(RawSpecMethod.create(item))
        return api_methods

    @staticmethod
    def parse_from_url(url: str) -> APIMethods:
        response = requests.get(url)

        if response.status_code != 200:
            raise ValueError(f"Failed to fetch data from {url}: status is {response.status_code}")

        return Parser.parse_from_raw(response.text)

    @staticmethod
    def parse_from_file(file_path: str) -> APIMethods:
        try:
            with open(file_path) as f:
                content = f.read()
            return Parser.parse_from_raw(content)
        except FileNotFoundError:
            raise ValueError(f"Failed to open file {file_path}: file not found")



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Set

from .models import APIMethod, APIMethods


@dataclass
class ContentDiff:
    content: str
    added_fields: List[str]
    missed_fields: List[str]

    def __init__(self, content: str):
        self.content = content
        self.added_fields = []
        self.missed_fields = []

    def set_added_fields(self, fields: List[str]) -> None:
        self.added_fields = fields

    def set_missed_fields(self, fields: List[str]) -> None:
        self.missed_fields = fields


@dataclass
class ResponseBodyDiff:
    code: str
    content_diff: Dict[str, ContentDiff]

    def __init__(self, code: str):
        self.code = code
        self.content_diff = {}

    def add_added_fields(self, content: str, fields: List[str]) -> None:
        if content not in self.content_diff and fields:
            self.content_diff[content] = ContentDiff(content)
        if fields:
            self.content_diff[content].set_added_fields(fields)

    def add_missed_fields(self, content: str, fields: List[str]) -> None:
        if content not in self.content_diff and fields:
            self.content_diff[content] = ContentDiff(content)
        if fields:
            self.content_diff[content].set_missed_fields(fields)


@dataclass
class APIMethodDiff:
    method: str
    route: str

    added_query_params: List[str]
    missed_query_params: List[str]

    request_body_diff: Dict[str, ContentDiff]

    added_status_codes: List[str]
    missed_status_codes: List[str]

    response_body_diff: Dict[str, ResponseBodyDiff]

    @classmethod
    def create(cls, method_data: APIMethod) -> "APIMethodDiff":
        return cls(
            method=method_data.method,
            route=method_data.route,
            added_query_params=[],
            missed_query_params=[],
            request_body_diff={},
            added_status_codes=[],
            missed_status_codes=[],
            response_body_diff={},
        )

    def set_added_query_params(self, query_params: Set[str]) -> None:
        self.added_query_params = sorted(query_params)

    def set_missed_query_params(self, query_params: Set[str]) -> None:
        self.missed_query_params = sorted(query_params)

    def add_added_request_body_fields(self, content: str, fields: List[str]) -> None:
        if content not in self.request_body_diff and fields:
            self.request_body_diff[content] = ContentDiff(content)
        if fields:
            self.request_body_diff[content].set_added_fields(fields)

    def add_missed_request_body_fields(self, content: str, fields: List[str]) -> None:
        if content not in self.request_body_diff and fields:
            self.request_body_diff[content] = ContentDiff(content)
        if fields:
            self.request_body_diff[content].set_missed_fields(fields)

    def set_added_status_codes(self, codes: Set[str]) -> None:
        self.added_status_codes = sorted(codes)

    def set_missed_status_codes(self, codes: Set[str]) -> None:
        self.missed_status_codes = sorted(codes)

    def add_added_response_body_fields(self, code: str, content: str, fields: List[str]) -> None:
        if code not in self.response_body_diff and fields:
            self.response_body_diff[code] = ResponseBodyDiff(code)
        if fields:
            self.response_body_diff[code].add_added_fields(content, fields)

    def add_missed_response_body_fields(self, code: str, content: str, fields: List[str]) -> None:
        if code not in self.response_body_diff and fields:
            self.response_body_diff[code] = ResponseBodyDiff(code)
        if fields:
            self.response_body_diff[code].add_missed_fields(content, fields)

    def is_modified(self) -> bool:
        if (
            self.added_query_params
            or self.missed_query_params
            or self.request_body_diff
            or self.added_status_codes
            or self.missed_status_codes
            or self.response_body_diff
        ):
            return True
        return False


@dataclass
class Diff:
    added: List[APIMethodDiff]
    similar: List[APIMethodDiff]
    modified: List[APIMethodDiff]
    missed: List[APIMethodDiff]

    def __init__(self) -> None:
        self.added: List[APIMethodDiff] = []
        self.similar: List[APIMethodDiff] = []
        self.modified: List[APIMethodDiff] = []
        self.missed: List[APIMethodDiff] = []

    def add_added(self, method: APIMethodDiff) -> None:
        self.added.append(method)

    def add_similar(self, method: APIMethodDiff) -> None:
        self.similar.append(method)

    def add_modified(self, method: APIMethodDiff) -> None:
        self.modified.append(method)

    def add_missed(self, method: APIMethodDiff) -> None:
        self.missed.append(method)


class Differ:
    def __init__(self, golden: APIMethods, testing: APIMethods) -> None:
        self.golden_ams = golden
        self.testing_ams = testing
        self.diff = Diff()

    @abstractmethod
    def get_diff(self) -> Any:
        pass

    def get_added_method_ids(self) -> List[str]:
        return sorted(self.golden_ams.get_ids() - self.testing_ams.get_ids())

    def get_missed_method_ids(self) -> List[str]:
        return sorted(self.testing_ams.get_ids() - self.golden_ams.get_ids())

    def get_common_method_ids(self) -> List[str]:
        return sorted(self.golden_ams.get_ids() & self.testing_ams.get_ids())

    def set_added_status_codes(self, method_id: str, method_diff: APIMethodDiff) -> None:
        method_diff.set_added_status_codes(
            self.golden_ams[method_id].get_codes() - self.testing_ams[method_id].get_codes()
        )

    def set_missed_status_codes(self, method_id: str, method_diff: APIMethodDiff) -> None:
        method_diff.set_missed_status_codes(
            self.testing_ams[method_id].get_codes() - self.golden_ams[method_id].get_codes()
        )

    def get_common_response_codes(self, method_id: str) -> List[str]:
        return sorted(self.golden_ams[method_id].get_codes() & self.testing_ams[method_id].get_codes())

    def set_added_queries(self, method_id: str, method_diff: APIMethodDiff) -> None:
        method_diff.set_added_query_params(
            self.golden_ams[method_id].query_params - self.testing_ams.methods[method_id].query_params
        )

    def set_missed_queries(self, method_id: str, method_diff: APIMethodDiff) -> None:
        method_diff.set_missed_query_params(
            self.testing_ams[method_id].query_params - self.golden_ams[method_id].query_params
        )

    def set_added_request_body_fields(
            self, method_id: str, method_diff: APIMethodDiff, content_type: str = "application/json"
    ) -> None:
        method_diff.add_added_request_body_fields(
            content_type,
            self.compare_schemas(
                self.golden_ams[method_id].request_body_schema[content_type].get("properties", {}),
                self.testing_ams[method_id].request_body_schema[content_type].get("properties", {})
            )
        )

    def set_missed_request_body_fields(
            self, method_id: str, method_diff: APIMethodDiff, content_type: str = "application/json"
    ) -> None:
        method_diff.add_missed_request_body_fields(
            content_type,
            self.compare_schemas(
                self.testing_ams.methods[method_id].request_body_schema[content_type].get("properties", {}),
                self.golden_ams.methods[method_id].request_body_schema[content_type].get("properties", {})
            )
        )

    def set_added_response_body_fields(
            self, method_id: str, method_diff: APIMethodDiff, code: str, content_type: str = "application/json"
    ) -> None:
        method_diff.add_added_response_body_fields(
            code,
            content_type,
            self.compare_schemas(
                self.golden_ams[method_id].response_body_schema[code][content_type].get("properties", {}),
                self.testing_ams[method_id].response_body_schema[code][content_type].get("properties", {})
            )
        )

    def set_missed_response_body_fields(
            self, method_id: str, method_diff: APIMethodDiff, code: str, content_type: str = "application/json"
    ) -> None:
        method_diff.add_missed_response_body_fields(
            code,
            content_type,
            self.compare_schemas(
                self.testing_ams[method_id].response_body_schema[code][content_type].get("properties", {}),
                self.golden_ams[method_id].response_body_schema[code][content_type].get("properties", {})
            )
        )

    def compare_schemas(
            self, golden_schema: Dict[str, Any], testing_schema: Dict[str, Any], path: str = ""
    ) -> List[str]:
        differences = []
        for key in golden_schema:
            current_path = f"{path}.{key}" if path else key
            if key not in testing_schema:
                differences.append(current_path)
            elif golden_schema[key]["type"] == "array" and golden_schema[key]["items"]["type"] == "object":
                differences.extend(self.compare_schemas(
                    golden_schema[key]["items"].get("properties", {}),
                    testing_schema[key]["items"].get("properties", {}),
                    current_path + ".[*]"
                ))
            elif golden_schema[key]["type"] == 'object':
                differences.extend(self.compare_schemas(
                    golden_schema[key].get("properties", {}),
                    testing_schema[key].get("properties", {}),
                    current_path
                ))
        return differences


@dataclass
class DiffDataCoverage:
    def __init__(self, diff: Diff):
        self.all: int = len(diff.added) + len(diff.similar) + len(diff.modified)
        self.full: int = len(diff.similar)
        self.partial: int = len(diff.modified)
        self.empty: int = len(diff.added)
        self.methods_full: List[APIMethodDiff] = diff.similar
        self.methods_partial: List[APIMethodDiff] = diff.modified
        self.methods_empty: List[APIMethodDiff] = diff.added
        self.full_percent: float = round(self.full / self.all * 100, 2)
        self.partial_percent: float = 100 - self.full_percent - round(self.empty / self.all * 100, 2)
        self.empty_percent: float = 100 - self.full_percent - self.partial_percent
        stat_min_percent = 5
        self.stat_full_percent: float = (
                100
                - (max(self.empty_percent, stat_min_percent) if self.empty_percent else 0)
                - (max(self.partial_percent, stat_min_percent) if self.partial_percent else 0)
        )
        self.stat_partial_percent: float = (
                100
                - self.stat_full_percent
                - (max(self.empty_percent, stat_min_percent) if self.empty_percent else 0)
        )
        self.stat_empty_percent: float = 100 - self.stat_full_percent - self.stat_partial_percent


class DifferCoverage(Differ):
    def get_diff(self) -> DiffDataCoverage:
        for method_id in self.get_added_method_ids():
            self.diff.add_added(APIMethodDiff.create(self.golden_ams[method_id]))

        for method_id in self.get_common_method_ids():
            method_diff = APIMethodDiff.create(self.golden_ams[method_id])

            self.set_added_queries(method_id, method_diff)
            self.set_added_request_body_fields(method_id, method_diff)
            self.set_added_status_codes(method_id, method_diff)

            for code in self.get_common_response_codes(method_id):
                self.set_added_response_body_fields(method_id, method_diff, code)

            if method_diff.is_modified():
                self.diff.add_modified(method_diff)
                continue

            self.diff.add_similar(method_diff)

        return DiffDataCoverage(self.diff)


class DiffDataDiscrepancy:
    def __init__(self, diff: Diff):
        self.methods_partial: List[APIMethodDiff] = diff.modified
        self.methods_empty: List[APIMethodDiff] = diff.missed


class DifferDiscrepancy(Differ):
    def get_diff(self) -> DiffDataDiscrepancy:
        for method_id in self.get_missed_method_ids():
            self.diff.add_missed(APIMethodDiff.create(self.testing_ams[method_id]))

        for method_id in self.get_common_method_ids():
            method_diff = APIMethodDiff.create(self.golden_ams[method_id])

            self.set_missed_queries(method_id, method_diff)
            self.set_missed_request_body_fields(method_id, method_diff)
            self.set_missed_status_codes(method_id, method_diff)

            for code in self.get_common_response_codes(method_id):
                self.set_missed_response_body_fields(method_id, method_diff, code)

            if method_diff.is_modified():
                self.diff.add_modified(method_diff)
                continue

        return DiffDataDiscrepancy(self.diff)


class DiffDataChanges:
    def __init__(self, diff: Diff):
        self.added: List[APIMethodDiff] = diff.added
        self.modified: List[APIMethodDiff] = diff.modified
        self.deleted: List[APIMethodDiff] = diff.missed


class DifferChanges(Differ):
    def get_diff(self) -> DiffDataChanges:
        for method_id in self.get_added_method_ids():
            self.diff.add_added(APIMethodDiff.create(self.golden_ams[method_id]))

        for method_id in self.get_missed_method_ids():
            self.diff.add_missed(APIMethodDiff.create(self.testing_ams[method_id]))

        for method_id in self.get_common_method_ids():
            method_diff = APIMethodDiff.create(self.golden_ams[method_id])

            self.set_added_queries(method_id, method_diff)
            self.set_missed_queries(method_id, method_diff)

            self.set_added_request_body_fields(method_id, method_diff)
            self.set_missed_request_body_fields(method_id, method_diff)

            self.set_added_status_codes(method_id, method_diff)
            self.set_missed_status_codes(method_id, method_diff)

            for code in self.get_common_response_codes(method_id):
                self.set_added_response_body_fields(method_id, method_diff, code)
                self.set_missed_response_body_fields(method_id, method_diff, code)

            if method_diff.is_modified():
                self.diff.add_modified(method_diff)
                continue

        return DiffDataChanges(self.diff)



import os
from typing import Any

from jinja2 import Environment, FileSystemLoader, Template

from .differ import DiffDataChanges, DiffDataCoverage, DiffDataDiscrepancy


class Generator:
    _PATH_TEMPLATES = os.path.dirname(os.path.realpath(__file__)) + '/templates'
    _TEMPLATE_COVERAGE = 'coverage.html.j2'
    _TEMPLATE_DISCREPANCY = 'discrepancy.html.j2'
    _TEMPLATE_CHANGES = 'changes.html.j2'

    def __init__(self) -> None:
        self._templates = Environment(loader=FileSystemLoader(self._PATH_TEMPLATES))

    def _get_template(self, template_name: str) -> Template:
        return self._templates.get_template(name=template_name)

    def _render_template(self, template_name: str, **kwargs: Any) -> str:
        template = self._get_template(template_name=template_name)
        return template.render(**kwargs)

    @staticmethod
    def _create_dir(dirname: str) -> None:
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

    def _create_file(self, file_path: str, content: str) -> None:
        self._create_dir(os.path.dirname(file_path))
        with open(file_path, 'w') as file:
            file.write(content)

    def render_coverage_report(self, diff: DiffDataCoverage) -> str:
        return self._render_template(self._TEMPLATE_COVERAGE, **vars(diff))

    def coverage_report(self, diff: DiffDataCoverage, file_path: str) -> None:
        content = self.render_coverage_report(diff)
        self._create_file(file_path, content)

    def render_discrepancy_report(self, diff: DiffDataDiscrepancy) -> str:
        return self._render_template(self._TEMPLATE_DISCREPANCY, **vars(diff))

    def discrepancy_report(self, diff: DiffDataDiscrepancy, file_path: str) -> None:
        content = self.render_discrepancy_report(diff)
        self._create_file(file_path, content)

    def render_changes_report(self, diff: DiffDataChanges) -> str:
        return self._render_template(self._TEMPLATE_CHANGES, **vars(diff))

    def changes_report(self, diff: DiffDataChanges, file_path: str) -> None:
        content = self.render_changes_report(diff)
        self._create_file(file_path, content)



import logging
from typing import Any

from .differ import DifferChanges
from .generator import Generator
from .parser import Parser


def changes(args: Any) -> None:
    logging.basicConfig(level=logging.INFO, format='%(message)s')
    logger = logging.getLogger("Changes")

    try:
        logger.info("Determination changes to the Open API spec")

        logger.info(f"Parsing the current spec: {args.current_spec_path}")
        current_spec_method = Parser.parse(args.current_spec_path)

        logger.info(f"Parsing the previous spec: {args.previous_spec_path}")
        previous_spec_method = Parser.parse(args.previous_spec_path)

        logger.info("Defining the difference")
        diff = DifferChanges(current_spec_method, previous_spec_method).get_diff()

        logger.info(f"Generating the changes report: {args.report_path}")
        Generator().changes_report(diff, args.report_path)

    except ValueError as e:
        logger.critical(f"{e}")



import argparse

from .changes import changes
from .coverage import coverage
from .discrepancy import discrepancy


def command() -> None:
    parser = argparse.ArgumentParser(description='vedro-spec-compare commands')
    subparsers = parser.add_subparsers(help='Available commands', required=True)

    coverage_parser = subparsers.add_parser('coverage', help='Generate coverage report')
    coverage_parser.add_argument('golden_spec_path', type=str, help='Path to the golden OpenAPI spec')
    coverage_parser.add_argument('testing_spec_path', type=str, help='Path to the testing OpenAPI spec')
    coverage_parser.add_argument(
        '--report-path', type=str, help='The path of the coverage report', default='coverage.html'
    )
    coverage_parser.set_defaults(func=coverage)

    discrepancy_parser = subparsers.add_parser('discrepancy', help='Generate discrepancy report')
    discrepancy_parser.add_argument('golden_spec_path', type=str, help='Path to the golden OpenAPI spec')
    discrepancy_parser.add_argument('testing_spec_path', type=str, help='Path to the testing OpenAPI spec')
    discrepancy_parser.add_argument(
        '--report-path', type=str, help='The path of the discrepancy report', default='discrepancy.html'
    )
    discrepancy_parser.set_defaults(func=discrepancy)

    changes_parser = subparsers.add_parser('changes', help='Generate changes report')
    changes_parser.add_argument('current_spec_path', type=str, help='Path to the current OpenAPI spec')
    changes_parser.add_argument('previous_spec_path', type=str, help='Path to the previous OpenAPI spec')
    changes_parser.add_argument(
        '--report-path', type=str, help='The path of the changes report', default='changes.html'
    )
    changes_parser.set_defaults(func=changes)

    args = parser.parse_args()
    args.func(args)
